/* ++++++++++
	file: driver.c

	info: based upon driver.c 'A skeletal device driver'
	         This driver implements the BJL communication protocols with
	         the Atari Juagar BJL developement System by 42Bastian Schick.
	         See readme.txt for more details on Jaguar developement.

	driver author: Robert Jurziga, (moveq@gmx.net)

	modifications: Roberto Nadal Martinez (swapd0@yahoo.es)
		Improved interface to enable/disable check sum, and now you can
		send/receive data.
+++++ */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// I N C L U D E S
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
#include <SupportDefs.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ByteOrder.h>
#include "../bjldriver.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// D E F I N E S
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
#define	BJL_DEVICE_NAME	"misc/bjlServer"
#define PACKET_SIZE 4096



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// P R O T O T Y P E S
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
void	SendLong	 	(uint32 value);
uint32	SendByte 		(uint8 value);

bool	ReceiveBool();
uint8	ReceiveByte();
uint32	ReceiveLong();

void PutByte(uint8 byte);

void WaitForBusyLow();
void WaitForBusyHigh();
uint8 IsBusy();

bool Strobe();
void SetStrobe(bool state);


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// D A T A
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
io_control_data_t config;
bigtime_t	waiter = 1;
uint32 port;
uint32 timeout;

uint8 byte;
uint8 busy, strobe;
uint8 fail;		// true if send/receive failed by timeout

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
status_t
init_hardware (void)
{
	return B_OK;
}



/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
/*
status_t
init_driver (void)
{
	return B_OK;
}
*/

/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
/*
void
uninit_driver (void)
{
}
*/
	

/* ----------
	my_device_open - handle open() calls
----- */
static status_t my_device_open (const char *name, uint32 flags
	, void** cookie)
{
	*cookie = NULL;
	return B_OK;
}


/* ----------
	my_device_read - handle read() calls
----- */

static status_t my_device_read (void* cookie, off_t position, void *buf
	, size_t* num_bytes)
{
	*(uint8 *)buf=ReceiveByte();
	if (!fail)
	{
		*num_bytes = 1;
		return B_OK;
	}
	else
	{
		*num_bytes=0;
		return B_ERROR;
	}
}


/* ----------
	Sends the buffer of size num_bytes to the Jaguar
----- */

static status_t my_device_write (void* cookie, off_t position, const void* buffer
	, size_t* num_bytes)
{
	int i;
	uint32 data;
	int32 packetSize,cksum = 0;
	uint32 *pbuffer=(uint32 *)buffer;

	config.numPackets=0.0;
	config.numFails=0.0;
//	set_thread_priority ( find_thread(NULL), B_LOW_PRIORITY	);

	// old mode
	if (config.checkSum && !config.packets)
		*num_bytes-=4;

	while (*num_bytes)
	{
		packetSize=min_c(PACKET_SIZE, *num_bytes);
		for ( i=0; i<packetSize; i+=4 )
		{
			snooze(waiter);
			data=B_BENDIAN_TO_HOST_INT32(*pbuffer++);
			cksum-=data;
			SendLong(data);
		}

		snooze(waiter);
		++config.numPackets;
		*num_bytes-=packetSize;	// ok
		if (config.checkSum)
		{
			SendLong(cksum);
			cksum=0;
			if (config.packets && !ReceiveBool())
			{
			// fail, move back pointer and add packet size
				*num_bytes+=packetSize;
				pbuffer-=(packetSize >> 2);
				++config.numFails;
			}
		}
	}

	return B_OK;
}


/* ---------- ---------- ---------- ---------- ---------- ---------- ---------- ---------- ---------- ----------
	my_device_control - handle ioctl calls
----- */

static status_t my_device_control(void* cookie, uint32 op, void* arg, size_t len)
{
	io_control_data_t *icd;

	icd=(io_control_data_t *)arg;			// cast pointer
	switch (op)
	{
		case CONFIGURE:
			config.port=icd->port;				// set all values
			port=config.port;
			config.checkSum=icd->checkSum;
			config.packets=icd->packets;
			config.timeout=icd->timeout;
			break;
		case SET_PORT:
			config.port=icd->port;				// set the port address
			port=config.port;
			break;
		case SET_CHECK_SUM:						// enable/disable checksum
			config.checkSum=icd->checkSum;
			break;
		case SET_PACKETS:						// enable/disable packets
			config.packets=icd->packets;
			break;
		case SET_TIMEOUT:						// set timeout count
			config.timeout=icd->timeout;
			break;
		case READ_STATUS:
			icd->ok=ReceiveBool();
			break;
		case GET_ERRORS:						// read errors from last write
			icd->numPackets=config.numPackets;
			icd->numFails=config.numFails;
			break;
		default:
			return (B_BAD_VALUE);
	}

	return (B_OK);
}


/* ----------
	my_device_close - handle close() calls
----- */

static status_t
my_device_close (void* cookie)
{
	PutByte(0);
	return B_OK;
}


/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
my_device_free (void* cookie)
{
	return B_OK;
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	BJL_DEVICE_NAME,
	NULL
};

/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	my_device_open, 			/* -> open entry point */
	my_device_close, 			/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	return my_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}



// ********
// The following low-level routines were written by 42Bastian Schick (asm)
// C version by swap d0
//  ******************************************************************
//
/*
int SendPacket(uint32 *buffer, int bytes)
{
	int32 cksum=0;
	for ( i=0; i<bytes; i+=4 )
	{
		snooze(waiter);
		data=B_BENDIAN_TO_HOST_INT32(*buffer++);
		cksum-=data;
		SendLong(data);
	}

	if (config.checkSum)
		SendLong(cksum);

	return cksum;
}
*/
void SendLong(uint32 value)
{
	SendByte(value);
	SendByte(value >> 8);
	SendByte(value >> 16);
	SendByte(value >> 24);
}

uint32 SendByte(uint8 value)
{
	WaitForBusyLow();
	PutByte(value);
	SetStrobe(false);
	WaitForBusyHigh();
	SetStrobe(true);
}

// ********
// This one by swap d0
//  ******************************************************************
//
uint8 ReceiveByte()
{
	uint8 data=0;
	int i;

	timeout=config.timeout;
	fail=0;

	SetStrobe(true);
	WaitForBusyLow();

	if (!fail)
		for ( i=0; i<8; ++i )
		{
			SetStrobe(false);

			snooze(10);
			data=data << 1;
			if (!IsBusy())
				data=data | 0x01;

			snooze(20);
			SetStrobe(true);
			snooze(30);
		}

	return data;
}

// reads one bit
bool ReceiveBool()
{
	bool data;

	timeout=config.timeout;
	fail=0;

	SetStrobe(true);
	WaitForBusyLow();

	if (!fail)
	{
		SetStrobe(false);
		snooze(20);
		if (!IsBusy())
			data=true;
		else
			data=false;

		SetStrobe(true);
		snooze(30);

//	WaitForBusyLow();
//	WaitForBusyHigh();
	}

	return data;
}

// ********
// This one by swap d0
//  ******************************************************************
//

void PutByte(uint8 byte)
{
	__asm__("
	movl port,%edx
	movl 0x8(%ebp),%eax
	outb %al,%dx			// write port value
	");
}

void WaitForBusyLow()
{
	__asm__("
	movl port, %edx			// numero del puerto
//	movl timeout, %ebx		// read timeout count
	incw %dx				// status port

waitBusyLow:
//	dec %ebx
//	jz l01					// timeout
	inb	%dx,%al				// al byte readed
	testb %al,%al
	jns waitBusyLow			// not signed -> /busy low
//	movl %ebx,timeout		// write timeout counter
//	xor %ax,%ax
//	jmp l02
//l01:
//	mov $0xffff,%ax
//l02:
//	movb %al,fail
	");
}

void WaitForBusyHigh()
{
	__asm__("
	movl port, %edx		// numero del puerto
//	movl timeout, %ebx		// read timeout count
	incw %dx				// status port

waitBusyHigh:
//	dec %ebx
//	jz l03					// timeout
	inb	%dx,%al				// al byte leido
	testb %al,%al
	js waitBusyHigh			// signed -> /busy high
//	movl %ebx,timeout		// write timeout counter
//	xor %ax,%ax
//	jmp l04
//l03:
//	mov $0xffff,%ax
//l04:
//	movb %al,fail
	");
}

uint8 IsBusy()
{
	__asm__("
	movl port, %edx		// numero del puerto
	incw %dx				// status port

	inb	%dx,%al				// al byte leido
	testb %al,%al
	js isHigh				// signed -> /busy high
	xor %ax,%ax
	jmp l05
isHigh:
	mov $0xffff,%ax
l05:

	movb %al,busy
	");

	return busy;
}

bool Strobe()
{
	static bool strobe;
	
	__asm__("
	movl port, %edx		// numero del puerto
	incw %dx				// status port
	incw %dx				// control port

	inb %dx,%al
	movb %al,strobe
	");

	return strobe;
}

void SetStrobe(bool state)
{
	if (state)
	{
		__asm__("
		movl port, %edx		// numero del puerto
		incw %dx				// status port
		incw %dx				// control port

		xor %al,%al
		outb %al,%dx			// 0 => 1 low active
		");
	}
	else
	{
		__asm__("
		movl port, %edx		// numero del puerto
		incw %dx				// status port
		incw %dx				// control port

		movb $1,%al
		outb %al,%dx			// 1 => 0 low active
		");
	}
}
