// Remote debugger
// Atari Jaguar side

	.68000

cReset			equ 'BOOM'		// reset Jaguar

cMotorolaState	equ 'MOTS'		// Motorola State
cGPUState		equ 'GPUS'		// GPU State
cDSPState		equ 'DSPS'		// DSP State

cMotorolaStep	equ 'MSTP'		// Motorola Step
cGPUStep		equ 'GSTP'		// GPU Step
cDSPStep		equ 'DSTP'		// DSP Step

cMotorolaSkip	equ 'MSKP'		// Motorola Skip one instruction
cGPUSkip		equ 'GSKP'		// GPU Skip one instruction
cDSPSkip		equ 'DSKP'		// DSP Skip one instruction

cMotorolaRun	equ 'MRUN'		// Motorola Run
cGPURun			equ 'GRUN'		// GPU Run
cDSPRun			equ 'DRUN'		// DSP Run

cMotorolaBreakPoint	equ 'MBRK'	// Motorola break point + address
cGPUBreakPoint	equ 'GBRK'		// GPU break point + address
cDSPBreakPoint	equ 'DBRK'		// DSP break point + address

cMemoryRead		equ 'MEMR'		// Memory read + address

cRestart		equ 'RSTR'		// Restart program

STATUS_SIZE		equ 8*2*4+4+2

RED		equ $f0ff
GREEN	equ $8fff
BLUE	equ $00ff
BLACK	equ $0000
WHITE	equ $88ff

MAX_BREAK_POINTS	equ 16
	rsreset
_address	rs.l 1
_opcode		rs.w 1
_breakPointSize rscount

	include "Jaguar.inc"

// initialize debugger:
//	writes into trace vector(9) $24 -> to trace routine
//	writes into trap #0 vector (32+n) $80 -> start the debugger
//	writes into illegal vector(4) $10 -> to break point routine
InitDebugger
	movem.l d0/a0,-(sp)
	move.l #TraceVector,$24.w		// trace vector
	move.l #StartDebugger,$80.w		// trap #0, jump into debug mode
	move.l $10.w,oldIllegal
	move.l #BreakPoint,$10.w		// illegal -> to implement break points
	lea breakPoints(pc),a0
	move.w #MAX_BREAK_POINTS-1,d0
.loop
	clr.l (a0)+						// clear address
	clr.w (a0)+						// clear opcode
	dbf d0,.loop
	movem.l (sp)+,d0/a0
	rts

// look for PC into break point list
// if don't exist
//	it's an illegal instruction -> executes old illegal vector (reset)
// if exist
//	write memory with old opcode and call TraceOn
_pc
	dc.l 0
BreakPoint
	move.l 2(sp),_pc
	movem.l d0-7/a0,-(sp)
	move.l _pc(pc),d0				// pc
	lea breakPoints(pc),a0
	move.w numBreakPoints,d7
	subq.w #1,d7
.loop
	cmp.l _address(a0),d0		// break point address = pc?
	beq.s .found				// yes
	lea _breakPointSize(a0),a0	// no, inc pointer in breakPointSize
	dbf d7,.loop
// pc not found into break point list -> it's an illegal instruction
	movea.l oldIllegal,a0
	jmp (a0)

.found
	move.w _opcode(a0),d1		// old opcode
	movea.l d0,a0				// pc address
	move.w d1,(a0)				// restore old pc
	movem.l (sp)+,d0-7/a0
	jmp TraceOn

// d0=command
ExecuteCommand
	lea commands(pc),a0
.find_next
	move.l (a0)+,d1			; command
	beq.s .not_found		; =0 => end of list
	cmp.l d0,d1				; d0 = d1 => command found
	beq.s .found
	lea 4(a0),a0			; skip command address
	bra.s .find_next

.not_found
	move.w #RED,BG
	rts

.found
	move.w #GREEN,BG
	movea.l (a0),a0
	jmp (a0)

// jump table of debugger commands
commands
// state commands
	dc.l cMotorolaState,MotorolaState
	dc.l cGPUState,GPUState
	dc.l cDSPState,DSPState
// step commands
	dc.l cMotorolaStep,MotorolaStep
	dc.l cGPUStep,GPUStep
	dc.l cDSPStep,DSPStep
// skip commands
	dc.l cMotorolaSkip,MotorolaSkip
	dc.l cGPUSkip,GPUSkip
	dc.l cDSPSkip,DSPSkip
// run commands
	dc.l cMotorolaRun,MotorolaRun
	dc.l cGPURun,GPURun
	dc.l cDSPRun,DSPRun
// break point commands
	dc.l cMotorolaBreakPoint,MotorolaBreakPoint
	dc.l cGPUBreakPoint,GPUBreakPoint
	dc.l cDSPBreakPoint,DSPBreakPoint
// memory commands
	dc.l cMemoryRead,MemoryRead
// misc commands
	dc.l cRestart,Restart
	dc.l 0,0						// end of list

// Executes motorola commands
MotorolaState
	lea status(pc),a0
	move.w #STATUS_SIZE-1,d0
	bsr SendBuffer
	rts

MotorolaStep
	movem.l status,d0-7/a0-7		// restore status
	move.l status_pc,2(sp)
	jmp TraceOn

MotorolaSkip
	bsr ReadCommand					// d0=next pc
	move.l d0,status_pc
	rts

MotorolaRun
	movem.l status,d0-7/a0-7		// restore status
	move.l status_pc,2(sp)
	andi.w #$7fff,(sp)
	rte

MotorolaBreakPoint
	bsr ReadCommand					// d0=address of break point
	lea breakPoints(pc),a0
	move.w numBreakPoints(pc),d7
.loop
	tst.w d7
	beq.s .end

	cmp.l _address(a0),d0		// break point address = d0?
	beq.s .found				// yes
	lea _breakPointSize(a0),a0	// no, inc pointer in breakPointSize
	subq.w #1,d7
	bra.s .loop
.end
	move.l d0,_address(a0)		// save break point address
	movea.l d0,a1
	move.w (a1),_opcode(a0)		// save opcode
	addq.w #1,numBreakPoints
	rts

.found							// remove break point
	movea.l d0,a1
	move.l _opcode(a0),(a1)
	lea _breakPointSize(a0),a1	// move list up
.copyLoop
	move.l (a1)+,(a0)+
	move.w (a1)+,(a0)+
	dbf d7,.copyLoop
	subq.w #1,numBreakPoints
	rts

// Executes GPU commands
GPUState
	bsr ReadGPUState
	lea riscRegs(pc),a0
	move.w #4*(32+4)-1,d0		// size in bytes
	bsr SendBuffer
	rts

SINGLE_STOP	equ SINGLE_STEP

GPUStep
	movea.l G_PC,a0
	lea 4(a0),a0				// +2 = correct pc + 2 = next instruction
	move.w (a0),d2				// d2 = next opcode
	move.l G_CTRL,d0
	or.w #SINGLE_STEP | RISCGO,d0
	move.l d0,G_CTRL			// do step

.waitDone
	move.l G_CTRL,d0
	and.w #SINGLE_STOP,d0
	beq.s .waitDone

	moveq #0,d0			// index
	moveq #10,d1		// shift
	lsr.w d1,d2			// d2 = opcode
.findNext
	move.w harzardStart(pc,d0.w),d3
	beq.s .end
	cmp.w d2,d3
	beq.s TraceHarzard
	addq.w #4,d0
	bra.s .findNext
.end
	rts

// opcode, size
harzardStart
	dc.w 38,6		// MOVEI
	dc.w 54,2		// MMULT
	dc.w 18,2		// IMULT
	dc.w 19,2		// RESMAC
	dc.w 20,2		// IMACN
	dc.w 52,2		// JUMP
	dc.w 53,2		// JR
	dc.w 0,0

harzardTable
	dc.w 38,6		// MOVEI
	dc.w 54,2		// MMULT
	dc.w 19,2		// RESMAC
	dc.w 20,2		// IMACN
	dc.w 52,2		// JUMP
	dc.w 53,2		// JR
	dc.w 0,0

TraceHarzard
	moveq #0,d7						// instructions to execute
	cmp.w #52,d2					// JUMP
	bne.s .no_jump
	moveq #1,d7
.no_jump

	addq.w #2,d0
	adda.w harzardTable(pc,d0.w),a0	// skip harzard instruction

.harzardBlock
	moveq #0,d0						// index
	move.w (a0),d2					// d1 = next opcode
	lsr.w d1,d2						// d2 = opcode
.findNext
	move.w harzardTable(pc,d0.w),d3
	beq.s .execute
	cmp.w d2,d3
	beq.s .harzardFound
	addq.w #4,d0
	bra.s .findNext
.harzardFound
	adda.w 2+harzardTable(pc,d0.w),a0	// next instruction
	addq.w #1,d7					// count it
	bra.s .harzardBlock

// executes d7 instructions
.execute
	move.l G_CTRL,d0
	or.w #SINGLE_GO,d0			// enable single go
	move.l d0,G_CTRL			// do step

.waitDone2
	move.l G_CTRL,d0
	and.w #SINGLE_STOP,d0		// wait until risc stopped
	beq.s .waitDone2
	dbf d7,.execute
	rts

GPUSkip
	rts

GPURun
	move.l #1,G_CTRL			// go!
	rts

GPUBreakPoint
	rts

ReadGPUState
// save GPU state
	move.l G_PC,riscPC					// save PC
	move.l G_FLAGS,riscFlags			// save flags
	move.l G_CTRL,riscControl			// save risc control
	move.l G_DIVCTRL,riscRemain			// divide remainder

// stop gpu
	move.l G_CTRL,d0
	and.w #$fffe,d0
	move.l d0,G_CTRL
.wait1
	move.l G_CTRL,d0
	btst #0,d0
	bne.s .wait1

// executes GPU code to save registers
	move.l #saveGPU,G_PC
	move.l G_CTRL,d0
	and.w #$ffe0,d0
	or.w #RISCGO,d0
	move.l d0,G_CTRL

// wait until finish
.wait2
	move.l G_CTRL,d0
	btst #0,d0
	bne.s .wait2

// restore GPU state
	addq.l #2,riscPC
	move.l riscPC,G_PC				// restore pc
	move.l riscFlags,G_FLAGS		// restore flags
	rts

// Executes DSP commands

StopGPU
	lea G_CTRL,a0
	moveq #-2,d0
	and.l (a0),d0
	move.l d0,(a0)
	rts

DSPState
	rts

DSPStep
	move.l D_CTRL,d0
	and.l #$ffffffe0,d0		// mask out control
	tst.w firstDSP
	bne.w .no_first
	or.w #$9,d0
	bra.s .do_it
.no_first
	or.w #$11,d0
.do_it
	move.l d0,D_CTRL		// do step
	rts

DSPSkip
	rts

DSPRun
	rts

DSPBreakPoint
	rts

MemoryRead
	bsr ReadCommand
	movea.l d0,a0			// d0 = address
	move.w #256-1,d0
	bsr SendBuffer
	rts

// Restart program
Restart
	movem.l status,d0-7/a0-7		// restore status
	move.l startAddress,2(sp)
	bra.s TraceOn

// return command/value in d0
ReadCommand
.timeout
	bsr ReceiveByte
	bcs.s .timeout
	move.b d0,d1
	ror.l #8,d1

	bsr ReceiveByte
	bcs.s .timeout
	move.b d0,d1
	ror.l #8,d1

	bsr ReceiveByte
	bcs.s .timeout
	move.b d0,d1
	ror.l #8,d1

	bsr ReceiveByte
	bcs.s .timeout
	move.b d0,d1
	ror.l #8,d1
	move.l d1,d0
	rts

// start the debugger, save interrupt vector and disable it
// put processor into trace mode
StartDebugger
	move.w VI,oldVI					// save vi counter
	move.w #$7fff,VI				// desable vi
//	move.l LEVEL0.w,

// warning continues on TraceOn

// executes one instruction
TraceOn
	ori.w #$8000,(sp)				// enable trace
	rte

// trace vector, after executing one instruction this routine is called
TraceVector
	movem.l d0-7/a0-7,status		// save status
	move.l 2(sp),status+8*2*4		// pc
	move.w (sp),status+8*2*4+4		// sr
.loop
	bsr ReadCommand					// d0=command
	bsr ExecuteCommand
	bra.s .loop

	include "Remote.inc"

	align phrase
saveGPU
	.gpu
	movei #riscRegs,r30
	store r0,(r30)
	addqt #4,r30
	store r1,(r30)
	addqt #4,r30
	store r2,(r30)
	addqt #4,r30
	store r3,(r30)
	addqt #4,r30
	store r4,(r30)
	addqt #4,r30
	store r5,(r30)
	addqt #4,r30
	store r6,(r30)
	addqt #4,r30
	store r7,(r30)
	addqt #4,r30
	store r8,(r30)
	addqt #4,r30
	store r9,(r30)
	addqt #4,r30
	store r10,(r30)
	addqt #4,r30
	store r11,(r30)
	addqt #4,r30
	store r12,(r30)
	addqt #4,r30
	store r13,(r30)
	addqt #4,r30
	store r14,(r30)
	addqt #4,r30
	store r15,(r30)
	addqt #4,r30
	store r16,(r30)
	addqt #4,r30
	store r17,(r30)
	addqt #4,r30
	store r18,(r30)
	addqt #4,r30
	store r19,(r30)
	addqt #4,r30
	store r20,(r30)
	addqt #4,r30
	store r21,(r30)
	addqt #4,r30
	store r22,(r30)
	addqt #4,r30
	store r23,(r30)
	addqt #4,r30
	store r24,(r30)
	addqt #4,r30
	store r25,(r30)
	addqt #4,r30
	store r26,(r30)
	addqt #4,r30
	store r27,(r30)
	addqt #4,r30
	store r28,(r30)
	addqt #4,r30
	store r29,(r30)
	addqt #4,r30
	store r30,(r30)			// r30...
	addqt #4,r30
	store r31,(r30)
	addqt #8,r30			// skip pc

	moveta r29,r29			// save r29
	movei #G_CTRL,r30
	load (r30),r29
	bclr #0,r29
	bset #3,r29
	store r29,(r30)
	movefa r29,r29			// restore r29
	nop
	nop
	nop
.hang
	nop
	nop
	nop
	jr t,.hang
	nop

	.68000
	even
oldIllegal
	dc.l 0		// old illegal vector

startAddress
	dc.l 0		// program start address

firstGPU
	dc.w 0
firstDSP
	dc.w 0
oldVI
	dc.w 0

	align long
riscRegs
	dc.l  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15
	dc.l 16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31
riscPC
	dc.l 0
riscFlags
	dc.l 0
riscControl
	dc.l 0
riscRemain
	dc.l 0

status
	ds.l 8*2	// data registers, address registers
status_pc
	ds.l 1		// pc
	ds.w 1		// sr

numBreakPoints
	ds.w 1		// break points number < MAX_BREAK_POINTS
breakPoints
	ds.b _breakPointSize*MAX_BREAK_POINTS	// break points array
	even

